/* v-hwclock -  Linux Hardware Clock utility

    Linux hardware clock manipulating utility
*/

/*

Copyright Notice:

Copyright (C) 2013-2014 R Veera Kumar

This file is part of v-lnx-tools.

v-lnx-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

v-lnx-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with v-lnx-tools.  If not, see <http://www.gnu.org/licenses/>.

See the COPYING file distributed together for license information.

See the AUTHOR file distributed together for authors information.

*/

/* USAGE: v-hwclock [ -h ] | [ -V ] | [ -r ] | [ -g ] | [ -s ] | [ -w ] | [ -u ] | [ -l ] | [ -f FILE ] */

#include "v-hwclock.h"

int main ( int argc, char *argv[] )

 {

  int cnt, r, g, s, w, u, l, f;
  r = g = s = w = u = l = f = 0;
  char *rfn = "/dev/rtc";
  int fdin, iocs;
  struct rtc_time rdrtc;
  struct timeval tv;
  struct timezone tz;
  struct tm brkt, brkttmp;
  time_t val1, t_t_p_1 = 100000l;
  long int time_var1;

#ifdef MUSL_ADDN
  struct timespec ts;
#endif

  if ( ( argc > 5 ) )
   { usage(); exit(1); }

  cnt = 1;
  while ( cnt <= (argc-1) )
   {

    if ( strcmp(argv[cnt], "-f") == 0 )
     if ( f != 1 )
      { f = 1; if ( (cnt+1) > (argc-1) ) { opt_arg_err(); exit(1); }; rfn = argv[cnt+1]; cnt +=2; continue; }
     else { opt_err(); usage(); exit(1); }

    else if ( strcmp(argv[cnt], "-r") == 0 ) { r = 1; ++cnt; }
    else if ( strcmp(argv[cnt], "-g") == 0 ) { g = 1; ++cnt; }
    else if ( strcmp(argv[cnt], "-s") == 0 ) { s = 1; ++cnt; }
    else if ( strcmp(argv[cnt], "-w") == 0 ) { w = 1; ++cnt; }
    else if ( strcmp(argv[cnt], "-u") == 0 ) { u = 1; ++cnt; }
    else if ( strcmp(argv[cnt], "-l") == 0 ) { l = 1; ++cnt; }

    else if ( strcmp(argv[cnt], "-h") == 0 )
     { usage(); exit(0); }

    else if ( strcmp(argv[cnt], "-V") == 0 )
     { version(); exit(0); }

    else { parm_unknown_err(argv[cnt]); usage(); exit(1); }

   }

  if ( (fdin = open(rfn, O_RDONLY)) == -1 ) { file_open_err(); exit(1); }

/* Start commands */

  if ( r == 1 )

   {

    if ( ( iocs = ioctl ( fdin, RTC_RD_TIME, &rdrtc ) ) == -1 ) { ioctl_err(); exit(1);  }

    printf ("%02d-%02d-%04d %02d:%02d:%02d\n", rdrtc.tm_mday, rdrtc.tm_mon + 1, rdrtc.tm_year + 1900, rdrtc.tm_hour, rdrtc.tm_min, rdrtc.tm_sec );
   }

  if ( g == 1 )

   {

    if ( gettimeofday (&tv, &tz) != 0 ) { gettimeofday_err() ; exit(1); }

    printf ("Seconds since epoch: %ld Microseconds: %ld\n", tv.tv_sec, tv.tv_usec );

    printf ("Timezone MinutesWest: %d DST: %d\n", tz.tz_minuteswest, tz.tz_dsttime );

    putenv("TZ=UTC-0");

    if ( localtime_r (&tv.tv_sec, &brkttmp) != &brkttmp ) { print_err("Unable to execute localtime_r"); exit (1); }

    printf ("%02d-%02d-%04d %02d:%02d:%02d UTC-0\n", brkttmp.tm_mday, brkttmp.tm_mon + 1, brkttmp.tm_year + 1900, brkttmp.tm_hour, brkttmp.tm_min, brkttmp.tm_sec);

#ifdef MUSL_ADDN
    clock_gettime(CLOCK_REALTIME, &ts);

    if ( localtime_r (&ts.tv_sec, &brkttmp) != &brkttmp ) { print_err("Unable to execute localtime_r"); exit (1); }

    printf ("%02d-%02d-%04d %02d:%02d:%02d UTC-0\n", brkttmp.tm_mday, brkttmp.tm_mon + 1, brkttmp.tm_year + 1900, brkttmp.tm_hour, brkttmp.tm_min, brkttmp.tm_sec);
#endif

   }

  if ( s == 1 )

   {

    if ( (l == 1) && ( u != 1 ) )

     {

      if ( localtime_r (&t_t_p_1, &brkttmp) != &brkttmp ) { print_err("Unable to execute localtime_r"); exit (1); }

      if ( ( iocs = ioctl ( fdin, RTC_RD_TIME, &rdrtc ) ) == -1 ) { ioctl_err(); exit(1);  }

      brkt.tm_sec = rdrtc.tm_sec;
      brkt.tm_min= rdrtc.tm_min;
      brkt.tm_hour = rdrtc.tm_hour;
      brkt.tm_mday = rdrtc.tm_mday;
      brkt.tm_mon = rdrtc.tm_mon;
      brkt.tm_year = rdrtc.tm_year;
      brkt.tm_gmtoff = brkttmp.tm_gmtoff;
      brkt.tm_zone = brkttmp.tm_zone;

      val1 =  mktime ( &brkt );

      if ( val1 == -1 ) { print_err("Unsuccessful mktime command"); exit(1); }

#ifdef DEBUG
      printf ("rtc2utcsec: %ld\n", val1);
#endif

      time_var1 = brkttmp.tm_gmtoff/60;

      tv.tv_sec = val1;
      tv.tv_usec = 0;
      tz.tz_minuteswest = (int) (-time_var1);
      tz.tz_dsttime = 0;

      if ( settimeofday (&tv, &tz ) != 0 ) { settimeofday_err() ; exit(1); }

     }

    if ( ( ( u == 1 ) && ( l != 1 ) ) || ( ( u != 1 ) && ( l != 1 )  ) )

     {

      putenv("TZ=UTC-0");
      if ( localtime_r (&t_t_p_1, &brkttmp) != &brkttmp ) { print_err("Unable to execute localtime_r"); exit (1); }

      if ( ( iocs = ioctl ( fdin, RTC_RD_TIME, &rdrtc ) ) == -1 ) { ioctl_err(); exit(1);  }

      brkt.tm_sec = rdrtc.tm_sec;
      brkt.tm_min= rdrtc.tm_min;
      brkt.tm_hour = rdrtc.tm_hour;
      brkt.tm_mday = rdrtc.tm_mday;
      brkt.tm_mon = rdrtc.tm_mon;
      brkt.tm_year = rdrtc.tm_year;
      brkt.tm_gmtoff = brkttmp.tm_gmtoff;
      brkt.tm_zone = brkttmp.tm_zone;

      val1 =  mktime ( &brkt );

      if ( val1 == -1 ) { print_err("Unsuccessful mktime command"); exit(1); }

#ifdef DEBUG
      printf ("rtc2utcsec: %ld\n", val1);
#endif

      time_var1 = brkttmp.tm_gmtoff/60;

      tv.tv_sec = val1;
      tv.tv_usec = 0;
      tz.tz_minuteswest = (int) (-time_var1);
      tz.tz_dsttime = 0;

      if ( settimeofday (&tv, &tz ) != 0 ) { settimeofday_err() ; exit(1); }

     }

   }

  if ( w == 1 )

   {

    if ( gettimeofday (&tv, &tz) != 0 ) { gettimeofday_err() ; exit(1); }

    if ( (l == 1) && ( u != 1 ) )

     {

      if ( localtime_r (&tv.tv_sec, &brkt) != &brkt ) { print_err("Unable to execute localtime_r"); exit (1); }

      rdrtc.tm_sec = brkt.tm_sec;
      rdrtc.tm_min = brkt.tm_min;
      rdrtc.tm_hour = brkt.tm_hour;
      rdrtc.tm_mday = brkt.tm_mday;
      rdrtc.tm_mon = brkt.tm_mon;
      rdrtc.tm_year = brkt.tm_year;
      rdrtc.tm_wday = 0;
      rdrtc.tm_yday = 0;
      rdrtc.tm_isdst = 0;
     
      if ( ( iocs = ioctl ( fdin, RTC_SET_TIME, &rdrtc ) ) == -1 ) { ioctl_err(); exit(1);  }

     }

    if ( ( ( u == 1 ) && ( l != 1 ) ) || ( ( u != 1 ) && ( l != 1 )  ) )

     {

      putenv("TZ=UTC-0");

      if ( localtime_r (&tv.tv_sec, &brkt) != &brkt ) { print_err("Unable to execute localtime_r"); exit (1); }

      rdrtc.tm_sec = brkt.tm_sec;
      rdrtc.tm_min = brkt.tm_min;
      rdrtc.tm_hour = brkt.tm_hour;
      rdrtc.tm_mday = brkt.tm_mday;
      rdrtc.tm_mon = brkt.tm_mon;
      rdrtc.tm_year = brkt.tm_year;
      rdrtc.tm_wday = 0;
      rdrtc.tm_yday = 0;
      rdrtc.tm_isdst = 0;
     
      if ( ( iocs = ioctl ( fdin, RTC_SET_TIME, &rdrtc ) ) == -1 ) { ioctl_err(); exit(1);  }

     }

   }

  close(fdin);

  return 0;

 }

void usage ( void )

 {
   fputs(DESC "\n", stderr);
   fputs("\nUsage:\n", stderr);
   fputs("\tv-hwclock  [ -h ] | [ -V ] | [ -r ] | [ -g ] | [ -s ] |\n", stderr);
   fputs("\t           [ -w ] | [ -u ] | [ -l ] | [ -f FILE ]\n", stderr);
   fputs("\n\tEach argument can be given only once.\n\n", stderr);
   fputs("\t-r\t Show hardware clock time.\n", stderr);
   fputs("\t-g\t Show system time.\n", stderr);
   fputs("\t-s\t Set system time from hardware clock.\n", stderr);
   fputs("\t-w\t Set hardware clock from system time.\n", stderr);
   fputs("\t-u\t Assume hardware clock is kept in UTC; (default). \n", stderr);
   fputs("\t-l\t Assume hardware clock is kept in local time.\n", stderr);
   fputs("\t-f FILE\t Use specified file as rtc (e.g. /dev/rtc0). \n", stderr);
   fputs("\t-h\t Show usage.\n", stderr);
   fputs("\t-V\t Show License and Version.\n", stderr);

 }

void version (void)

 {

  fputs(DESC "\n", stderr);
  fputs("Version: " VERSION "\n", stderr);
  fputs("License: " LICENSE "\n", stderr);

 }

void opt_err (void)
 {
  fputs("Each argument can be given only once!\n", stderr);
 }

void opt_arg_err (void)
 {
  fputs("Neccessary argument for specified option not given!\n", stderr);
 }

void file_open_err (void)
 {
  fputs("Could not open file for read/write!\n", stderr);
 }

void parm_unknown_err (char *arg)
 {
  fprintf(stderr, "Unknown argument given!: %s\n", arg);
 }

void file_read_err (void)
 {
  fputs("Read Error! Exiting...\n", stderr);
 }

void ioctl_err (void)
 {
  fputs("Error: Unable to execute ioctl! Exiting...\n", stderr);
 }

void gettimeofday_err (void)
 {
  fputs("Error: Unable to execute gettimeofday! Exiting...\n", stderr);
 }

void settimeofday_err (void)
 {
  fputs("Error: Unable to execute settimeofday! Exiting...\n", stderr);
 }

void print_err ( char * error_name)
 {
  fprintf(stderr, "Error: %s! Exiting...\n", error_name );
 }

