/* v-pwgen - A ascii password generator for unix/linux */

/*

Copyright Notice:

Copyright (C) 2013-2014 R Veera Kumar

This file is part of v-lnx-tools.

v-lnx-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

v-lnx-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with v-lnx-tools.  If not, see <http://www.gnu.org/licenses/>.

See the COPYING file distributed together for license information.

See the AUTHOR file distributed together for authors information.

*/

/* USAGE: v-pwgen -sa | -ca | -nm | -pc | -a | -l length | [ -f random_device ] | [ -al number ] | [ -ac STRING ] | [ -rc STRING ] | -h | -V */

#include "v-pwgen.h"

FILE *fp;

int main ( int argc, char **argv )
 {

  char *arg_len, *fn, * arg_aln, * arg_acs, * arg_rcs, * acs, * rcs;
  fn = "/dev/urandom";
  int arg_f, arg_l, arg_sa, arg_ca, arg_nm, arg_pc, arg_a, arg_al, arg_ac, arg_rc;
  arg_f = arg_l = arg_sa = arg_ca = arg_nm = arg_pc = arg_a = arg_al = arg_ac = arg_rc = 0;
  int cnt, byte, i, j, k, st;
  int id_n, id_c, id_s, id_p;
  id_n = id_c = id_s = id_p = 0;

  unsigned int length, aln;
  unsigned int altc, tmp;
  uint8_t * mem, * memidx, round = 0;
  size_t len_acs, len_rcs;

  if ( argc > 15 )
   { usage(); exit(1); }

  cnt = 1;
  while ( cnt <= (argc-1) )
   {

    if ( strcmp(argv[cnt], "-l") == 0 )
     {
      if ( arg_l != 1 )
       { arg_l = 1; if ( (cnt+1) > (argc-1) ) { print_simple_error(err_no_assoc_arg); usage(); exit(1); }; arg_len = argv[cnt+1]; cnt +=2; continue; }
      else { print_simple_error(err_opt_once_only); usage(); exit(1); }
     }

    else if ( strcmp(argv[cnt], "-f") == 0 )
     {
      if ( arg_f != 1 )
       { arg_f = 1; if ( (cnt+1) > (argc-1) ) { print_simple_error(err_no_assoc_arg); usage(); exit(1); }; fn = argv[cnt+1]; cnt +=2; continue; }
      else { print_simple_error(err_opt_once_only); usage(); exit(1); }
     }

    else if ( strcmp(argv[cnt], "-al") == 0 )
     {
      if ( arg_al != 1 )
       { arg_al = 1; if ( (cnt+1) > (argc-1) ) { print_simple_error(err_no_assoc_arg); usage(); exit(1); }; arg_aln = argv[cnt+1]; cnt +=2; continue; }
      else { print_simple_error(err_opt_once_only); usage(); exit(1); }
     }

    else if ( strcmp(argv[cnt], "-ac") == 0 )
     {
      if ( arg_ac != 1 )
       { arg_ac = 1; if ( (cnt+1) > (argc-1) ) { print_simple_error(err_no_assoc_arg); usage(); exit(1); }; arg_acs = argv[cnt+1]; cnt +=2; continue; }
      else { print_simple_error(err_opt_once_only); usage(); exit(1); }
     }

    else if ( strcmp(argv[cnt], "-rc") == 0 )
     {
      if ( arg_rc != 1 )
       { arg_rc = 1; if ( (cnt+1) > (argc-1) ) { print_simple_error(err_no_assoc_arg); usage(); exit(1); }; arg_rcs = argv[cnt+1]; cnt +=2; continue; }
      else { print_simple_error(err_opt_once_only); usage(); exit(1); }
     }

    else if ( strcmp(argv[cnt], "-sa") == 0 )
     {
      if ( arg_sa != 1 ) { arg_sa = 1; ++cnt; }
      else { print_simple_error(err_opt_once_only); usage(); exit(1); }
     }
    else if ( strcmp(argv[cnt], "-ca") == 0 )
     {
      if ( arg_ca != 1 ) { arg_ca = 1; ++cnt; }
      else { print_simple_error(err_opt_once_only); usage(); exit(1); }
     }
    else if ( strcmp(argv[cnt], "-nm") == 0 )
     {
      if ( arg_nm != 1 ) { arg_nm = 1; ++cnt; }
      else { print_simple_error(err_opt_once_only); usage(); exit(1); }
     }
    else if ( strcmp(argv[cnt], "-pc") == 0 )
     {
      if ( arg_pc != 1 ) { arg_pc = 1; ++cnt; }
      else { print_simple_error(err_opt_once_only); usage(); exit(1); }
     }
    else if ( strcmp(argv[cnt], "-a") == 0 )
     {
      if ( arg_a != 1 ) { arg_a = 1; ++cnt; arg_sa = arg_ca = arg_nm = arg_pc = 1; }
      else { print_simple_error(err_opt_once_only); usage(); exit(1); }
     }
    else if ( strcmp(argv[cnt], "-h") == 0 ) { usage(); exit(0); }
    else if ( strcmp(argv[cnt], "-V") == 0 ) { version(); exit(0); }
    else { fprintf( stderr, "Error: Unknown argument given!: %s Exiting...\n\n", argv[cnt] ); usage(); exit(1); }

   }

  if ( arg_l != 1 ) { print_simple_error(err_len); usage(); exit(1); }

  length = (unsigned int) strtoul(arg_len, NULL, 10);

  if ( length >= 65537 ) { print_simple_error(err_len_too_much); usage(); exit(1); }

  if ( length < 1 ) { print_simple_error(err_len_lesser); usage(); exit(1); }

  if ( arg_al == 1 )
   {

    aln = (unsigned int) strtoul(arg_aln, NULL, 10);

    if ( aln > length ) { print_simple_error(err_atleast_too_much); usage(); exit(1); }

    if ( aln < 1 ) { print_simple_error(err_atleast_lesser); usage(); exit(1); }
   }

  mem = (uint8_t *) malloc ( (size_t) length );

  if ( mem == NULL ) { print_simple_error(err_mem_alloc_fail); exit(1); }

  memset ( mem, 0, 1 );

  if ( arg_ac )
   {
    len_acs = strlen(arg_acs);
    acs = (char *) malloc ( len_acs + 1);
    memcpy(acs, arg_acs, strlen(arg_acs)+1);
   }

  if ( arg_rc )
   {
    len_rcs = strlen(arg_rcs);
    rcs = (char *) malloc ( len_rcs + 1);
    memcpy(rcs, arg_rcs, strlen(arg_rcs)+1);
   }

  if ( mlock( mem, length ) != 0 ) { print_simple_error(err_mem_lock_fail); exit(1); }

  memidx = mem;

  fp = fopen ( fn, "r" );

  if ( fp ==NULL ) { fprintf ( stderr, "Error: Can't open file -> \"%s\" ! Exiting...\n", fn ); exit(1); }

  if ( (arg_sa == 1) || (arg_ca == 1) || (arg_nm == 1) || (arg_pc == 1) || (arg_ac == 1) )
   {
    if (arg_ac == 1)
     {
      str_uniq_char(acs);
     }

    if ( arg_rc == 1 )
     {
      str_uniq_char(rcs);
     }

    if ( (arg_ac == 1) && (arg_rc == 1) )
     {
      strs_uniq_char(acs, rcs);
     }

    if ( arg_ac ) len_acs = strlen(acs);
    if ( arg_rc ) len_rcs = strlen(rcs);

    if ( (arg_nm == 1) && (arg_rc == 1) )
     {
      for ( i = 0, k = 0; i < 10; ++i)
       {
        st = 0;
        for ( j = 0; (unsigned) j < len_rcs; ++j )
         {
          if ( ac[i] == (uint8_t) *(rcs+j) ) st = 1;
         }
        if ( st == 0 ) { act[k] = ac[i]; ++k; }
       }
      if ( k == 0 ) { arg_nm = 0; id_n = 0; }
      if ( k >= 1 ) { id_n = k; round += k; }
     }

    if ((arg_nm == 1) && (arg_rc != 1))
     {
      for ( i = 0; i < 10; ++i) act[i] = ac[i];
      id_n = 10;
      round += 10;
     }

    if ( (arg_nm != 1) && (arg_ac == 1) )
     {
      for ( i = 0, j = 0; (unsigned) i < len_acs; ++i )
       {
        if ( ((uint8_t) *(acs+i) >= 48) && ((uint8_t) *(acs+i) <= 57) )
         {
          act[j] = (uint8_t) *(acs+i);
          ++j;
         }
       }
      if ( j == 0 ) { arg_nm = 0; id_n = 0; }
      if ( j >= 1 ) { id_n = j; round += j; arg_nm = 1; }
     }

    if ( arg_nm == 0 ) id_n = 0;

    if ( (arg_ca == 1) && (arg_rc == 1) )
     {
      for ( i = 0, k = 0; i < 26; ++i)
       {
        st = 0;
        for ( j = 0; (unsigned) j < len_rcs; ++j )
         {
          if ( ac[i+10] == (uint8_t) *(rcs+j) ) st = 1;
         }
        if ( st == 0 ) { act[id_n+k] = ac[i+10]; ++k; }
       }
      if ( k == 0 ) { arg_ca = 0; id_c = 0; }
      if ( k >= 1 ) { id_c = k; round += k; }
     }

    if ((arg_ca == 1) && (arg_rc != 1))
     {
      for ( i = 0; i < 26; ++i) act[i+id_n] = ac[i+10];
      id_c = 26;
      round += 26;
     }

    if ( (arg_ca != 1) && (arg_ac == 1) )
     {
      for ( i = 0, j = 0; (unsigned) i < len_acs; ++i )
       {
        if ( ((uint8_t) *(acs+i) >= 65) && ((uint8_t) *(acs+i) <= 90) )
         {
          act[id_n+j] = (uint8_t) *(acs+i);
          ++j;
         }
       }
      if ( j == 0 ) { arg_ca = 0; id_c = 0; }
      if ( j >= 1 ) { id_c = j; round += j; arg_ca = 1; }
     }

    if ( arg_ca == 0 ) id_c = 0;

    if ( (arg_sa == 1) && (arg_rc == 1) )
     {
      for ( i = 0, k = 0; i < 26; ++i)
       {
        st = 0;
        for ( j = 0; (unsigned) j < len_rcs; ++j )
         {
          if ( ac[i+10+26] == (uint8_t) *(rcs+j) ) st = 1;
         }
        if ( st == 0 ) { act[id_n+id_c+k] = ac[i+10+26]; ++k; }
       }
      if ( k == 0 ) { arg_sa = 0; id_s = 0; }
      if ( k >= 1 ) { id_s = k; round += k; }
     }

    if ((arg_sa == 1) && (arg_rc != 1))
     {
      for ( i = 0; i < 26; ++i) act[i+id_n+id_c] = ac[i+10+26];
      id_s = 26;
      round += 26;
     }

    if ( (arg_sa != 1) && (arg_ac == 1) )
     {
      for ( i = 0, j = 0; (unsigned) i < len_acs; ++i )
       {
        if ( ((uint8_t) *(acs+i) >= 97) && ((uint8_t) *(acs+i) <= 122) )
         {
          act[id_n+id_c+j] = (uint8_t) *(acs+i);
          ++j;
         }
       }
      if ( j == 0 ) { arg_sa = 0; id_s = 0; }
      if ( j >= 1 ) { id_s = j; round += j; arg_sa = 1; }
     }

    if ( arg_sa == 0 ) id_s = 0;

    if ( (arg_pc == 1) && (arg_rc == 1) )
     {
      for ( i = 0, k = 0; i < 33; ++i)
       {
        st = 0;
        for ( j = 0; (unsigned) j < len_rcs; ++j )
         {
          if ( ac[i+10+26+26] == (uint8_t) *(rcs+j) ) st = 1;
         }
        if ( st == 0 ) { act[id_n+id_c+id_s+k] = ac[i+10+26+26]; ++k; }
       }
      if ( k == 0 ) { arg_pc = 0; id_p = 0; }
      if ( k >= 1 ) { id_p = k; round += k; }
     }

    if ((arg_pc == 1) && (arg_rc != 1))
     {
      for ( i = 0; i < 33; ++i) act[i+id_n+id_c+id_s] = ac[i+10+26+26];
      id_p = 33;
      round += 33;
     }

    if ( (arg_pc != 1) && (arg_ac == 1) )
     {
      for ( i = 0, j = 0; (unsigned) i < len_acs; ++i )
       {
        if (
            ( ((uint8_t) *(acs+i) >= 32) && ((uint8_t) *(acs+i) <= 47) ) || 
            ( ((uint8_t) *(acs+i) >= 58) && ((uint8_t) *(acs+i) <= 64) ) ||
            ( ((uint8_t) *(acs+i) >= 91) && ((uint8_t) *(acs+i) <= 96) ) ||
            ( ((uint8_t) *(acs+i) >= 123) && ((uint8_t) *(acs+i) <= 126) )
           )
         {
          act[id_n+id_c+id_s+j] = (uint8_t) *(acs+i);
          ++j;
         }
       }
      if ( j == 0 ) { arg_pc = 0; id_p = 0; }
      if ( j >= 1 ) { id_p = j; round += j; arg_pc = 1; }
     }

    if ( arg_pc == 0 ) id_p = 0;

   }

  if ( (arg_sa == 1) || (arg_ca == 1) || (arg_nm == 1) || (arg_pc == 1) )

   {

    cnt = length;

    if ( arg_al == 1 )
     {
      altc = 0;
      if ( arg_nm == 1 ) ++altc;
      if ( arg_sa == 1 ) ++altc;
      if ( arg_ca == 1 ) ++altc;
      if ( arg_pc == 1 ) ++altc;

      tmp = length / altc;
      if ( tmp )
       {
        if ( aln > tmp ) aln = tmp;

        if ( arg_nm == 1 )
         {
          for ( i = 0; (unsigned) i < aln; ++i )
           {
            byte = rndbyte(id_n);
            *memidx = act[byte+0]; ++memidx; --cnt;
           }
         }

        if ( arg_ca == 1 )
         {
          for ( i = 0; (unsigned) i < aln; ++i )
           {
            byte = rndbyte(id_c);
            *memidx = act[byte+id_n]; ++memidx; --cnt;
           }
         }
        if ( arg_sa == 1 )
         {
          for ( i = 0; (unsigned) i < aln; ++i )
           {
            byte = rndbyte(id_s);
            *memidx = act[byte+id_n+id_c]; ++memidx; --cnt;
           }
         }
        if ( arg_pc == 1 )
         {
          for ( i = 0; (unsigned) i < aln; ++i )
           {
            byte = rndbyte(id_p);
            *memidx = act[byte+id_n+id_c+id_s]; ++memidx; --cnt;
           }
         }
       }
     }

    while ( cnt != 0 )

     {

      byte = rndbyte(round);

      if ( arg_nm )
       {
        if ( byte < id_n ) { *memidx = act[byte+0]; ++memidx; --cnt; continue; }
        else byte -= id_n;
       }
      if ( arg_ca )
       {
        if ( byte < id_c ) { *memidx = act[byte+id_n]; ++memidx; --cnt; continue; }
        else byte -= id_c;
       }
      if ( arg_sa )
       {
        if ( byte < id_s ) { *memidx = act[byte+id_n+id_c]; ++memidx; --cnt; continue; }
        else byte -= id_s;
       }
      if ( arg_pc )
       {
        if ( byte < id_p ) { *memidx = act[byte+id_n+id_c+id_s]; ++memidx; --cnt; continue; }
        else byte -= id_p;
       }

     }

   }

  i = length;

  if ( length <=256 )
   {
    while ( i > 1 )
     {
      --i;
      j = rndbyte(i);
      k = *(mem+i);
      *(mem+i) = *(mem +j);
      *(mem + j ) = k;
     }
   }
  if ( length > 256 )
   {
    while ( i > 1 )
     {
      --i;
      j = rndword(i);
      k = *(mem+i);
      *(mem+i) = *(mem +j);
      *(mem + j ) = k;
     }
   }

  for ( i = 0; (unsigned) i < length; ++i )
   {
    if ( *(mem+i) == 0) break;
    putc(*(mem+i), stdout);
   }

  fflush(NULL);

  memset ( mem , 0, length );
  munlock ( mem, length );
  free(mem);

  if ( arg_ac ) free(acs);
  if ( arg_rc ) free(rcs);

  fclose ( fp );

  return 0;

 }

uint8_t rndbyte (uint8_t endrng)

 {

  uint8_t stat, round;
  int rndbyte;

  round = (255 / endrng) * endrng;

  stat = 0;

  while ( stat != 1 )
   {  
    rndbyte = getc(fp);
    if (rndbyte == EOF ) continue;
    if ( rndbyte > round ) continue;
    rndbyte = (rndbyte % endrng);
    stat = 1;
   }

  return (uint8_t) rndbyte;

 }

uint16_t rndword (uint16_t endrng)

 {

  uint16_t stat, round, rndword;
  int byte;

  round = (65536 / endrng) * endrng;

  stat = 0;

  while ( stat != 1 )
   {  
    byte = getc(fp);
    if (byte == EOF ) continue;
    rndword = 0; rndword = (uint16_t) byte; rndword <<= 8;
    while ( ( byte = getc(fp) ) == EOF  );
    rndword |= (uint16_t) byte;
    if ( rndword > round ) continue;
    rndword = (rndword % endrng);
    stat = 1;
   }

  return rndword;

 }

void str_uniq_char ( char * str )

 {

  int i, j, len;
  char tmp;

  len = strlen(str);

  for ( i = 0; i < len; ++i )
   {
    for ( j = i + 1; j < len; ++j )
     {
      if ( (unsigned char) *(str+i) == (unsigned char) *(str+j) ) { *(str+j) = (unsigned char) 0xff; }
     }
   }

  i = j = 0;
  while ( *(str+i) != '\0' )
   {
    if ( (unsigned char) *(str+i) != 0xff ) { tmp = (unsigned char) *(str+i); *(str+j) = (unsigned char) tmp; ++j; }
    ++i;
   }
  *(str+j) = '\0';

 }

void strs_uniq_char ( char * str1, char * str2 )

 {

  int i, j, len1, len2;
  char tmp;

  len1 = strlen(str1);
  len2 = strlen(str2);

  for ( i = 0; i < len1; ++i)
   {
    for ( j = 0; j < len2; ++j )
     {
      tmp = (unsigned char) *(str1+i);
      if ( (unsigned char) *(str1+i) == 0xff ) break;
      if ( (unsigned char) tmp == (unsigned char) *(str2+j) )
       {
        *(str1+i) = *(str2+j) = (unsigned char) 0xff;
       }
     }
   }

  i = j = 0;
  while ( *(str1+i) != '\0' )
   {
    if ( (unsigned char) *(str1+i) != 0xff ) { tmp = (unsigned char) *(str1+i); *(str1+j) = (unsigned char) tmp; ++j; }
    ++i;
   }
  *(str1+j) = '\0';

  i = j = 0;
  while ( *(str2+i) != '\0' )
   {
    if ( (unsigned char) *(str2+i) != 0xff ) { tmp = (unsigned char) *(str2+i); *(str2+j) = (unsigned char) tmp; ++j; }
    ++i;
   }
  *(str2+j) = '\0';

 }

void usage (void)
 {
  fputs(DESC "\n\n", stderr);
  fputs("Usage:\n", stderr);
  fputs("\tv-pwgen -sa | -ca | -nm | -pc | -a | -l length | [ -f random_device ] |\n", stderr);
  fputs("\t\t[ -al number ] | [ -ac STRING ] | [ -rc STRING ] | -h | -V\n", stderr);
  fputs("\n\tEach argument can be given only once.\n", stderr);
  fputs("\t-l\t Specify generated password length.\n", stderr);
  fputs("\t-f\t Specify random device. E.g. /dev/urandom.\n", stderr);
  fputs("\t-al\t Generate atleast numbers of each type.\n", stderr);
  fputs("\t-sa\t Generated password has small case alphabets.\n", stderr);
  fputs("\t-ca\t Generated password has capital case alphabets.\n", stderr);
  fputs("\t-nm\t Generated password has numericals.\n", stderr);
  fputs("\t-pc\t Generated password has punctuation characters.\n", stderr);
  fputs("\t-a\t Generated password has all the sa, ca, nm, pc characters.\n", stderr);
  fputs("\t-ac\t Generated password has characters from STRING.\n", stderr);
  fputs("\t-rc\t Generated password has no characters from STRING.\n", stderr);
  fputs("\t   \t -ac and -rc cancel each others characters.\n", stderr);
  fputs("\t-h\t Show help.\n", stderr);
  fputs("\t-V\t Show License and Version.\n", stderr);
 }

void version (void)
 {
  fputs(DESC"\n", stderr);
  fputs("Version: " VERSION "\n", stderr);
  fputs("License: " LICENSE "\n", stderr);
 }

void print_simple_error ( char * error )
 {
  fprintf(stderr, "Error: %s\n", error);
 }
