/* v-pwgen.h */

/*

Copyright Notice:

Copyright (C) 2013-2014 R Veera Kumar

This file is part of v-lnx-tools.

v-lnx-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

v-lnx-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with v-lnx-tools.  If not, see <http://www.gnu.org/licenses/>.

See the COPYING file distributed together for license information.

See the AUTHOR file distributed together for authors information.

*/

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <string.h>
#include <sys/mman.h>
#include <stddef.h>

void usage (void);
void version (void);

void print_simple_error ( char * error );

uint8_t rndbyte (uint8_t endrng);
uint16_t rndword (uint16_t endrng);
void str_uniq_char ( char * str );
void strs_uniq_char ( char * str1, char * str2 );

#define DESC "v-pwgen - A ascii password generator for unix/linux"
#define VERSION "0.1"
#define LICENSE "GNU GPL Version 3"

char * err_opt_once_only = "Each argument can be given only once!\n";
char * err_no_assoc_arg = "Neccessary argument for specified option not given!\n";
char * err_len = "Password length not given!\n";
char * err_len_too_much = "Password length is greater than 65536 bytes!\n";
char * err_len_lesser = "Password length is lesser than a byte!\n";
char * err_mem_alloc_fail = "Can't allocate memory!\n";
char * err_mem_lock_fail = "Can't lock memory!\n";
char * err_atleast_too_much = "Option -al value is greater than password length!\n";
char * err_atleast_lesser = "Option -al value is less than one!\n";

static const uint8_t ac[95] =
 {
  48, 49, 50, 51, 52, 53, 54, 55, 56, 57, // 0 - 9 ( ARRAY IND 0 - 9 )
  65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, // A - Z ( ARRAY IND 10 - 35)
  78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90,
  97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, // a - z ( ARRAY IND 36 - 61 )
  110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122,
  32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, // punctuation characters ( ARRAY IND 62 - 77 )
  58, 59, 60, 61, 62, 63, 64, // ( ARRAY IND 78 - 84 )
  91, 92, 93, 94, 95, 96, // ( ARRAY IND 85 - 90 )
  123, 124, 125, 126 // ( ARRAY IND 91 - 94
 };

static uint8_t act[95];

